import itertools

def flatten_args(*args):
    items = []
    for arg in args:
        if not isinstance(arg, type) and callable(arg):
            items.append(flatten_args(arg()))
        else:
            items.append(arg)
    return items


def on_exception(callback, *args):
    """Invoke callback if exception within args occurs. Reraise exception."""
    def decorator(main_function):
        def wrapper(*wargs, **kwargs):
            try:
                return main_function(*wargs, **kwargs)
            except Exception as e:
                # Extract from function if provided one.
                if len(args) and not isinstance(args[0], type) and callable(args[0]):
                    exceptions = args[0]()
                else:
                    exceptions = args
                # Determine action
                if len(exceptions) == 0 or isinstance(e, exceptions):
                    callback()
                raise e
        return wrapper
    return decorator


def on_returned(callback, *args):
    """Invoke callback if return is in args. Returns value."""
    def decorator(main_function):
        def wrapper(*wargs, **kwargs):
                returned = main_function(*wargs, **kwargs)
                # Extract from function if provided one.
                if len(args) and not isinstance(args[0], type) and callable(args[0]):
                    exceptions = args[0]()
                else:
                    exceptions = args
                # Determine action
                if len(exceptions) == 0 and returned is None:
                    callback()
                elif returned in (exceptions):
                    callback()
                return returned
        return wrapper
    return decorator



"""
    Filter Exception
If exception occurs, invoke callback and return value or allow exception 
to rise.

    Filter Return
If function returns value within args, invoke callback and return value or 
allow exception to rise.

    Filter
Allows for either exception or return to be the trigger for the callback.
"""

def filter_returned(callback, default=None, return_map={}):
    def decorator(main_function):
        def wrapper(*wargs, **kwargs):
            returned = main_function(*wargs, **kwargs)
            if returned in return_map.keys():
                return return_map[returned]
            else:
                return default
        return wrapper
    return decorator


def filter_exception(callback, default=None, exception_map={}):
    def decorator(main_function):
        def wrapper(*wargs, **kwargs):
            try:
                return main_function(*wargs, **kwargs)
            except Exception as e:
                if type(e) in exception_map.keys():
                    raise exception_map[type(e)]
                elif default is not None:
                    raise default
                else:
                    raise e
        return wrapper
    return decorator


"""
    Ensure Return
Ensures that the function returns a value of some kind.
"""
def default_return(default=None):
    def decorator(main_function):
        def wrapper(*wargs, **kwargs):
            returned = main_function(*wargs, **kwargs)
            return default if returned is None else returned
        return wrapper
    return decorator


def ensure_type(default=None, return_type=None):
    def decorator(main_function):
        def wrapper(*wargs, **kwargs):
            returned = main_function(*wargs, **kwargs)
            if return_type is None:
                return default if returned is None else returned
            return default if type(returned) == return_type else returned
        return wrapper
    return decorator


"""
Check return
"""
def check_return_type(types=[]):
    def decorator(main_function):
        def wrapper(*wargs, **kwargs):
            returned = main_function(*wargs, **kwargs)
            if type(returned) not in types:
                raise TypeError()
            return returned
        return wrapper
    return decorator



def on_raised(callback, *args):
    """Invoke callback if exception within args occurs. Reraise exception."""
    def decorator(main_function):
        def wrapper(*wargs, **kwargs):
            try:
                return main_function(*wargs, **kwargs)
            except Exception as e:
                if len(args) == 0 or isinstance(e, args):
                    callback(main_function, e)
                raise e
        return wrapper
    return decorator


def on_returned(callback, *args):
    """Invoke callback if return is in args. Returns value."""
    def decorator(main_function):
        def wrapper(*wargs, **kwargs):
                returned = main_function(*wargs, **kwargs)
                if len(args) == 0:
                    callback(main_function, returned)
                elif returned in (args):
                    callback(main_function, returned)
                return returned
        return wrapper
    return decorator


def no_raise(callback=lambda f, e: 0, instead_return=None):
    def decorator(main_function):
        def wrapper(*wargs, **kwargs):
            try:
                return main_function(*wargs, **kwargs)
            except Exception as e:
                callback(main_function, e)
                return instead_return
        return wrapper
    return decorator



from terse import on_raised
from terse import on_returned
from terse import no_raise


on_raised_counter = 0
def test_on_raised_none():
    global on_raised_counter
    # SETUP PREDICTABLE COUNTER FUNCTION
    def increment_counter(f, e):
        global on_raised_counter
        on_raised_counter += 1
    # SETUP PASS THROUGH FUNCTION
    @on_raised(increment_counter)
    def invoke_error(error):
        if isinstance(error, Exception):
            raise error
        else:
            return error
    # TEST PASS THROUGH
    on_raised_counter = 0
    assert invoke_error(0) == 0
    assert on_raised_counter == 0
    assert invoke_error(1) == 1
    assert on_raised_counter == 0
    assert invoke_error(2) == 2
    assert on_raised_counter == 0
    # TEST DEFAULT
    try:
        invoke_error(IOError())
        assert False
    except IOError:
        assert on_raised_counter == 1
    try:
        invoke_error(OSError())
        assert False
    except OSError:
        assert on_raised_counter == 2
    try:
        invoke_error(NotImplementedError())
        assert False
    except NotImplementedError:
        assert on_raised_counter == 3


def test_on_raised_singular():
    global on_raised_counter
    # SETUP PREDICTABLE COUNTER FUNCTION
    def increment_counter(f, r):
        global on_raised_counter
        on_raised_counter += 1
    # SETUP PASS THROUGH FUNCTION
    @on_raised(increment_counter, IndexError)
    def invoke_error_one(error):
        if isinstance(error, Exception):
            raise error
        else:
            return error
    # TEST PASS THROUGH
    on_raised_counter = 0
    assert invoke_error_one(0) == 0
    assert on_raised_counter == 0
    assert invoke_error_one(1) == 1
    assert on_raised_counter == 0
    assert invoke_error_one(2) == 2
    assert on_raised_counter == 0
    # TEST SINGULAR
    try:
        invoke_error_one(IndexError())
        assert False
    except IndexError:
        assert on_raised_counter == 1
    try:
        invoke_error_one(OSError())
        assert False
    except OSError:
        assert on_raised_counter == 1
    try:
        invoke_error_one(NotImplementedError())
        assert False
    except NotImplementedError:
        assert on_raised_counter == 1


def test_on_raised_many():
    global on_raised_counter
    # SETUP PREDICTABLE COUNTER FUNCTION
    def increment_counter(f, r):
        global on_raised_counter
        on_raised_counter += 1
    # SETUP PASS THROUGH FUNCTION
    @on_raised(increment_counter, IndexError, KeyError)
    def invoke_error_many(error):
        if isinstance(error, Exception):
            raise error
        else:
            return error
    # TEST PASS THROUGH
    on_raised_counter = 0
    assert invoke_error_many(0) == 0
    assert on_raised_counter == 0
    assert invoke_error_many(1) == 1
    assert on_raised_counter == 0
    assert invoke_error_many(2) == 2
    assert on_raised_counter == 0
    # TEST MULTIPLE
    try:
        invoke_error_many(IndexError())
        assert False
    except IndexError:
        assert on_raised_counter == 1
    try:
        invoke_error_many(KeyError())
        assert False
    except KeyError:
        assert on_raised_counter == 2
    try:
        invoke_error_many(NotImplementedError())
        assert False
    except NotImplementedError:
        assert on_raised_counter == 2


on_returned_counter = 0
def test_on_returned_empty():
    global on_returned_counter
    def increment_counter(f, r):
        global on_returned_counter
        on_returned_counter += 1
    @on_returned(increment_counter)
    def return_value(returned_value):
        return returned_value
    on_returned_counter = 0
    assert on_returned_counter == 0
    assert return_value(None) == None
    assert on_returned_counter == 1
    assert return_value(1) == 1
    assert on_returned_counter == 2
    assert return_value(2) == 2
    assert on_returned_counter == 3


def test_on_returned_singular():
    global on_returned_counter
    def increment_counter(f, r):
        global on_returned_counter
        on_returned_counter += 1
    @on_returned(increment_counter, 1)
    def return_value_one(returned_value):
        return returned_value
    on_returned_counter = 0
    assert on_returned_counter == 0
    assert return_value_one(1) == 1
    assert on_returned_counter == 1
    assert return_value_one(0) == 0
    assert on_returned_counter == 1
    assert return_value_one(1) == 1
    assert on_returned_counter == 2
    assert return_value_one(0) == 0
    assert on_returned_counter == 2


def test_on_returned_many():
    global on_returned_counter
    def increment_counter(f, r):
        global on_returned_counter
        on_returned_counter += 1
    @on_returned(increment_counter, 1, 2)
    def return_value_many(returned_value):
        return returned_value
    on_returned_counter = 0
    assert on_returned_counter == 0
    assert return_value_many(1) == 1
    assert on_returned_counter == 1
    assert return_value_many(2) == 2
    assert on_returned_counter == 2
    assert return_value_many(3) == 3
    assert on_returned_counter == 2
    assert return_value_many(4) == 4
    assert on_returned_counter == 2


def test_no_raise():
    @no_raise(instead_return=False)
    def f1():
        return True
    @no_raise(instead_return=False)
    def f2():
        raise Exception()
    @no_raise(instead_return=10)
    def f3():
        raise OSError()
    # RUN ALL THREE
    assert f1() == True
    assert f2() == False
    assert f3() == 10



import sys

def main(func):
    """Main Decorator"""
    def decorator(main_function):
        if main_function.__module__ == '__main__':
            returned = main_function()
            try:
                sys.exit(int(returned))
            except (TypeError, ValueError):
                sys.exit(0)
    return decorator



# Change Directory
from .in_directory import InDirectory
from .in_directory import in_dir
# Invoke
from .invoke import main
# On Exit
from .on_exit import on_raised
from .on_exit import on_returned
from .on_exit import no_raise



from .on_exit import on_exception
from .on_exit import on_returned
from .on_exit import filter_returned
from .on_exit import filter_exception
from .on_exit import default_return
from .on_exit import ensure_type
from .on_exit import check_return_type



import os


class InDirectory:
    def __init__(self, path):
        self.old_cwd = os.getcwd()
        self.path = path

    def __enter__(self):
        self.old_cwd = os.getcwd()
        try:
            os.chdir(self.path())
        except TypeError:
            os.chdir(self.path)

    def __exit__(self, exc_type, exc_val, exc_tb):
        os.chdir(self.old_cwd)


def in_dir(path):
    def decorator(fn):
        def wrapper(*args, **kwargs):
            with InDirectory(path):
                return fn(*args, **kwargs)
        return wrapper
    return decorator



import os
from terse import InDirectory
from terse import in_dir

#TODO Test lambda argument
def test_chdir():
    cwd = os.getcwd()
    with InDirectory('..'):
        assert cwd != os.getcwd()
    assert cwd == os.getcwd()
    with InDirectory(lambda: '..'):
        assert cwd != os.getcwd()
    assert cwd == os.getcwd()

#TODO Test lambda argument
def test_in_dir():
    cwd = os.getcwd()
    @in_dir('..')
    def func(ocwd, oval):
        assert ocwd != os.getcwd()
        return oval
    val = 123456789
    assert func(cwd, val) == val
    assert cwd == os.getcwd()
    @in_dir(lambda: '..')
    def func2(ocwd, oval):
        assert ocwd != os.getcwd()
        return oval
    val = 987654321
    assert func2(cwd, val) == val
    assert cwd == os.getcwd()



#!/usr/bin/env python3
from setuptools import setup, find_packages
from os import path
import pypandoc
proj_path = path.abspath(path.dirname(__file__))
md_path = path.join(proj_path, 'README.md')
long_description = pypandoc.convert(md_path, 'rst')

setup(
    author='Chad Lucas',
    author_email='cjlucas85@gmail.com',
    description='a collection of function decorators to handle common procedures done on the entry and exit points.',
    long_description=long_description,
    long_description_content_type='text/markdown',
    name='terse',
    url='https://github.com/cjlucas85/terse',
    version='0.0.8',
    packages=find_packages(),
    classifiers=(
        'Development Status :: 4 - Beta',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.5',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
    ),
)
